# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for jmp._src.loss_scale."""

import warnings

from absl.testing import absltest
from absl.testing import parameterized
import jax
import jax.numpy as jnp
from jmp._src import loss_scale as jmp
import numpy as np


class LossScaleTest(parameterized.TestCase):

  def test_no_op_loss_scale(self):
    loss_scale = jmp.NoOpLossScale()
    tree = {"a": jnp.ones([])}
    self.assertIs(loss_scale.scale(tree), tree)
    self.assertIs(loss_scale.unscale(tree), tree)

  @parameterized.named_parameters(
      ("StaticLossScale(2)", jmp.StaticLossScale, 2),
      ("StaticLossScale(3)", jmp.StaticLossScale, 3),
      ("StaticLossScale(4)", jmp.StaticLossScale, 4),
      ("DynamicLossScale(2)", jmp.DynamicLossScale, 2.),
      ("DynamicLossScale(3)", jmp.DynamicLossScale, 3.),
      ("DynamicLossScale(4)", jmp.DynamicLossScale, 4.),
  )
  def test_static_loss_scale(self, cls, scale):
    loss_scale = cls(scale)
    tree = {"a": jnp.array(1.)}
    scaled_tree = {"a": jnp.array(1. * scale)}
    self.assertEqual(loss_scale.scale(tree), scaled_tree)
    self.assertEqual(loss_scale.unscale(scaled_tree), tree)

  @parameterized.named_parameters(
      ("NoOpLossScale", jmp.NoOpLossScale),
      ("StaticLossScale", lambda: jmp.StaticLossScale(0)),
  )
  def test_static_empty_trees(self, create):
    loss_scale = create()
    self.assertEmpty(jax.tree_util.tree_leaves(loss_scale))

  def test_dynamic_loss_scale_no_warnings(self):
    with warnings.catch_warnings(record=True) as logged_warnings:
      jmp.DynamicLossScale(2. ** 15)
    self.assertEmpty(logged_warnings)

  def test_dynamic_loss_scale_tree(self):
    scale = jnp.ones([])
    counter = jnp.zeros([], jnp.int32)
    period = 2000
    factor = 2
    loss_scale = jmp.DynamicLossScale(scale, counter, period, factor)
    self.assertEqual(jax.tree_util.tree_leaves(loss_scale), [scale, counter])
    self.assertEqual(jax.tree_util.tree_map(lambda x: x, loss_scale),
                     loss_scale)

  @parameterized.parameters((20, 2), (30, 3))
  def test_dynamic_loss_scale_adjust_increases_on_finite(self, period, factor):
    grads_finite = jnp.bool_(True)
    loss_scale = jmp.DynamicLossScale(jnp.float32(10), jnp.int32(0),
                                      period, factor)
    for i in range(1, period):
      loss_scale = loss_scale.adjust(grads_finite)
      self.assertEqual(loss_scale.loss_scale, 10)
      self.assertEqual(loss_scale.counter, i)
      self.assertEqual(loss_scale.period, period)
      self.assertEqual(loss_scale.factor, factor)

    # Loss scale should wrap.
    loss_scale = loss_scale.adjust(grads_finite)
    self.assertEqual(loss_scale.loss_scale, 10 * factor)
    self.assertEqual(loss_scale.counter, 0)
    self.assertEqual(loss_scale.period, period)
    self.assertEqual(loss_scale.factor, factor)

  @parameterized.parameters((20, 2), (30, 3))
  def test_dynamic_loss_scale_adjust_reduce_on_non_finite(self, period, factor):
    grads_finite = jnp.bool_(False)
    init = np.float32(10)
    loss_scale = jmp.DynamicLossScale(jnp.asarray(init), jnp.int32(0), period,
                                      factor)
    self.assertLess(init / (factor ** 100), 1, msg="should cover max(1, S)")
    for i in range(100):
      loss_scale = loss_scale.adjust(grads_finite)
      np.testing.assert_allclose(loss_scale.loss_scale,
                                 max(1, init / (factor ** (i + 1))),
                                 rtol=1e-5)
      self.assertEqual(loss_scale.counter, 0)
      self.assertEqual(loss_scale.period, period)
      self.assertEqual(loss_scale.factor, factor)

  @parameterized.parameters((20, 2, .3125), (30, 3, .37), (5., 2., 0.))
  def test_dynamic_loss_scale_explicit_min_loss_scale(self, period, factor,
                                                      min_loss_scale):
    grads_finite = jnp.bool_(False)
    init = np.float32(10)
    loss_scale = jmp.DynamicLossScale(
        jnp.asarray(init), jnp.int32(0), period, factor,
        jnp.asarray(min_loss_scale))
    self.assertLess(init / (factor**100), 1, msg="should cover max(1, S)")
    for i in range(100):
      loss_scale = loss_scale.adjust(grads_finite)
      np.testing.assert_allclose(
          loss_scale.loss_scale,
          max(min_loss_scale, init / (factor**(i + 1))),
          rtol=1e-5)
      self.assertEqual(loss_scale.counter, 0)
      self.assertEqual(loss_scale.period, period)
      self.assertEqual(loss_scale.factor, factor)

  def test_dynamic_loss_scale_adjust_requires_scalar_input(self):
    pass

  def test_dynamic_loss_scale_raises_type_error_on_int_loss_scale(self):
    expected_message = "Expected floating type for loss_scale"
    with self.assertWarnsRegex(Warning, expected_message):
      jmp.DynamicLossScale(jnp.asarray(1, dtype=jnp.int32))

  def test_dynamic_loss_scale_raises_type_error_on_int_min_loss_scale(self):
    expected_message = "Expected floating type for min_loss_scale"
    with self.assertWarnsRegex(Warning, expected_message):
      jmp.DynamicLossScale(jnp.asarray(1, dtype=jnp.float32),
                           min_loss_scale=jnp.asarray(1, dtype=jnp.int32))

  @parameterized.parameters(jnp.inf, jnp.nan)
  def test_all_finite(self, non_finite):
    self.assertTrue(jmp.all_finite(None))
    self.assertTrue(jmp.all_finite({}))
    self.assertFalse(jmp.all_finite({"a": jnp.array(non_finite)}))
    self.assertFalse(jmp.all_finite({"a": jnp.ones([]),
                                     "b": jnp.array(non_finite)}))
    self.assertFalse(jmp.all_finite({"a": jnp.array(non_finite),
                                     "b": jnp.ones([])}))
    self.assertTrue(jmp.all_finite({"a": jnp.ones([]), "b": jnp.ones([])}))

  def test_select_tree(self):
    a = {"a": jnp.ones([]), "b": jnp.zeros([])}
    b = {"a": jnp.zeros([]), "b": jnp.ones([])}
    self.assertIsNone(jmp.select_tree(jnp.bool_(True), None, None))
    self.assertIsNone(jmp.select_tree(jnp.bool_(False), None, None))
    self.assertEqual(jmp.select_tree(jnp.bool_(True), a, b), a)
    self.assertEqual(jmp.select_tree(jnp.bool_(False), a, b), b)

  def test_select_tree_rejects_non_scalar(self):
    with self.assertRaisesRegex(AssertionError, "expected boolean scalar"):
      jmp.select_tree(jnp.ones([1]), None, None)

if __name__ == "__main__":
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for loss scaling."""

import dataclasses
import functools
from typing import Tuple, TypeVar, Union
import warnings

import jax
from jax import tree_util
import jax.numpy as jnp
import numpy as np

# from deepmind.internal import usage_logging

T = TypeVar("T")


def register_empty_pytree(cls):
  tree_util.register_pytree_node(cls, lambda x: ((), x), lambda x, _: x)


@dataclasses.dataclass(frozen=True)
class NoOpLossScale:
  """No-op loss scale does nothing."""

  @property
  def loss_scale(self):
    return 1

  def scale(self, tree: T) -> T:
    # usage_logging.log_event(usage_logging.Event.JMP, "NoOpLossScale")
    return tree

  def unscale(self, tree: T) -> T:
    return tree

  def adjust(self, grads_finite: jnp.ndarray):
    del grads_finite
    return self


@dataclasses.dataclass(frozen=True)
class StaticLossScale:
  """Scales and unscales by a fixed constant."""

  loss_scale: jnp.ndarray

  def scale(self, tree: T) -> T:
    # usage_logging.log_event(usage_logging.Event.JMP, "StaticLossScale")
    return jax.tree_util.tree_map(lambda x: x * self.loss_scale, tree)

  def unscale(self, tree: T) -> T:
    inv_loss_scale = 1 / self.loss_scale
    return jax.tree_util.tree_map(lambda x: x * inv_loss_scale, tree)

  def adjust(self, grads_finite: jnp.ndarray):
    del grads_finite
    return self

_Data = Tuple[jnp.ndarray, ...]
_Meta = Tuple[int, int]


@dataclasses.dataclass(frozen=True)
class DynamicLossScale:
  """Dynamic loss scale.

  Dynamic loss scaling tries to determine the largest loss scale value that
  will keep gradients finite. It does this by increasing the loss scale every
  `period` steps by `factor` if the grads remain finite, otherwise it reduces
  the loss scale by `1 / factor` and resets the counter.

      loss_scale = 2 ** 15
      counter = 0
      period = 2000
      factor = 2

      for step in range(num_steps):
        loss *= loss_scale
        grads /= loss_scale
        grads_finite = all_finite(grads)

        if grads_finite:
          counter += 1
          if counter == period:
            counter = 0
            loss_scale = first_finite(loss_scale * factor, loss_scale)
        else:
          counter = 0
          loss_scale = max(1, loss_scale / factor)

  Typical usage of this class will be something like:

  >>> loss_scale = jmp.DynamicLossScale(jnp.asarray(2. ** 15))
  >>> for _ in range(num_steps):
  ...   # compute loss
  ...   loss = loss_scale.scale(loss)
  ...   # compute grads
  ...   grads = loss_scale.unscale(grads)
  ...   grads_finite = jmp.all_finite(grads)
  ...   loss_scale = loss_scale.adjust(grads_finite)
  ...   # conditionally update params using grads
  """
  loss_scale: jnp.ndarray
  counter: jnp.ndarray = dataclasses.field(
      default_factory=lambda: np.zeros([], np.int32))
  period: int = 2000
  factor: int = 2
  min_loss_scale: jnp.ndarray = dataclasses.field(
      default_factory=lambda: np.ones([], np.float32))

  def __post_init__(self) -> None:
    warn_if_not_floating(self.loss_scale, "loss_scale")
    warn_if_not_floating(self.min_loss_scale, "min_loss_scale")

  def scale(self, tree: T) -> T:
    # usage_logging.log_event(usage_logging.Event.JMP, "DynamicLossScale")
    return jax.tree_util.tree_map(lambda x: x * self.loss_scale, tree)

  def unscale(self, tree: T) -> T:
    inv_loss_scale = 1 / self.loss_scale
    return jax.tree_util.tree_map(lambda x: x * inv_loss_scale, tree)

  def tree_flatten(self) -> Tuple[_Data, _Meta]:
    data = (self.loss_scale, self.counter)
    meta = (self.period, self.factor)
    return data, meta

  @classmethod
  def tree_unflatten(cls, meta: _Meta, data: _Data) -> "DynamicLossScale":
    loss_scale, counter = data
    period, factor = meta
    return cls(loss_scale, counter, period, factor)

  def adjust(self, grads_finite: jnp.ndarray) -> "DynamicLossScale":
    """Returns the next state dependent on whether grads are finite."""
    assert grads_finite.ndim == 0, "Expected boolean scalar"

    first_finite = lambda a, b: jax.lax.select(jnp.isfinite(a).all(), a, b)
    loss_scale = jax.lax.select(
        grads_finite,

        # When grads are finite increase loss scale periodically.
        jax.lax.select(
            self.counter == (self.period - 1),
            first_finite(self.loss_scale * self.factor,
                         self.loss_scale),
            self.loss_scale),

        # If grads are non finite reduce loss scale.
        jnp.maximum(self.min_loss_scale, self.loss_scale / self.factor))

    counter = ((self.counter + 1) % self.period) * grads_finite

    return DynamicLossScale(
        loss_scale=loss_scale,
        counter=counter,
        period=self.period,
        factor=self.factor,
        min_loss_scale=self.min_loss_scale)


register_empty_pytree(NoOpLossScale)
register_empty_pytree(StaticLossScale)
tree_util.register_pytree_node_class(DynamicLossScale)

LossScale = Union[NoOpLossScale, StaticLossScale, DynamicLossScale]


def all_finite(tree) -> jnp.ndarray:
  """Returns a scalar ndarray indicating whether the input arrays are finite."""
  leaves = jax.tree_util.tree_leaves(tree)
  if not leaves:
    return jnp.array(True)
  else:
    leaves = map(jnp.isfinite, leaves)
    leaves = map(jnp.all, leaves)
    return jnp.stack(list(leaves)).all()


def select_tree(pred: jnp.ndarray, a: T, b: T) -> T:
  """Selects a pytree based on the given predicate."""
  assert pred.ndim == 0 and pred.dtype == jnp.bool_, "expected boolean scalar"
  return jax.tree_util.tree_map(functools.partial(jax.lax.select, pred), a, b)


def warn_if_not_floating(x: Union[jnp.ndarray, object], var_name: str) -> None:
  """Produces a warning if the given array does not have a floating type.

  This function handles an edgecase where Jax passes in an `object()` to
  determine the structure of user defined pytrees during compilation. They
  recommend explicitly checking if the array in question has the type `object`.

  From the Jax documentation: "The __init__ and __new__ methods of custom
  PyTree classes should generally avoid doing any array conversion or other
  input validation, or else anticipate and handle these special cases."

  See:
  https://jax.readthedocs.io/en/latest/pytrees.html#custom-pytrees-and-initialization

  Args:
    x: Any object.
    var_name: A useful name to put in error messages.
  """
  if type(x) is object:  # pylint: disable=unidiomatic-typecheck
    return
  x_dtype = jax.eval_shape(lambda: x).dtype
  if not jnp.issubdtype(x_dtype, jnp.floating):
    warnings.warn(f"Expected floating type for {var_name}, got {x_dtype}")



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for mixed precision in JAX."""

import dataclasses
from typing import TypeVar

import jax
import jax.numpy as jnp
import numpy as np

T = TypeVar("T")


def _cast_floating_to(tree: T, dtype: jnp.dtype) -> T:
  def conditional_cast(x):
    if (isinstance(x, (np.ndarray, jnp.ndarray)) and
        jnp.issubdtype(x.dtype, jnp.floating)):
      x = x.astype(dtype)
    return x
  return jax.tree_util.tree_map(conditional_cast, tree)


@dataclasses.dataclass(frozen=True)
class Policy:
  """Encapsulates casting for inputs, outputs and parameters."""
  param_dtype: jnp.dtype
  compute_dtype: jnp.dtype
  output_dtype: jnp.dtype

  def cast_to_param(self, x):
    """Converts floating point values to the param dtype."""
    return _cast_floating_to(x, self.param_dtype)

  def cast_to_compute(self, x):
    """Converts floating point values to the compute dtype."""
    return _cast_floating_to(x, self.compute_dtype)

  def cast_to_output(self, x):
    """Converts floating point values to the output dtype."""
    return _cast_floating_to(x, self.output_dtype)

  def with_output_dtype(self, output_dtype: jnp.dtype) -> "Policy":
    return dataclasses.replace(self, output_dtype=output_dtype)

  def __str__(self):
    return "p={},c={},o={}".format(dtype_to_names[self.param_dtype][0],
                                   dtype_to_names[self.compute_dtype][0],
                                   dtype_to_names[self.output_dtype][0])


def get_policy(policy_name: str) -> Policy:
  """Returns a mixed precision policy parsed from a string."""
  # Loose grammar supporting:
  #  - "c=f16"         (params full, compute+output in f16),
  #  - "p=f16,c=f16"   (params, compute and output in f16).
  #  - "p=f16,c=bf16"  (params in f16, compute in bf16, output in bf16)
  # For values that are not specified params defaults to f32, compute follows
  # params and output follows compute (e.g. 'c=f16' -> 'p=f32,c=f16,o=f16').
  param_dtype = jnp.float32
  compute_dtype = output_dtype = None
  if "=" in policy_name:
    for part in policy_name.split(","):
      key, value = part.split("=", 2)
      value = parse_dtype(value)
      if key == "p" or key == "params":
        param_dtype = value
      elif key == "c" or key == "compute":
        compute_dtype = value
      elif key == "o" or key == "output":
        output_dtype = value
      else:
        raise ValueError(f"Unknown key '{key}' in '{policy_name}' should be "
                         "'params', 'compute' or 'output'.")
    if compute_dtype is None:
      compute_dtype = param_dtype
    if output_dtype is None:
      output_dtype = compute_dtype
  else:
    # Assume policy name is a dtype (e.g. 'f32' or 'half') that all components
    # of the policy should contain.
    param_dtype = compute_dtype = output_dtype = parse_dtype(policy_name)

  return Policy(param_dtype=param_dtype, compute_dtype=compute_dtype,
                output_dtype=output_dtype)


def cast_to_full(tree: T) -> T:
  """Ensures floating point leaves of the given tree are f32."""
  return _cast_floating_to(tree, jnp.float32)


def cast_to_half(tree: T) -> T:
  """Ensures floating point leaves of the given tree are half precision."""
  return _cast_floating_to(tree, half_dtype())


def half_dtype() -> jnp.dtype:
  """Returns the half precision dtype for the current backend."""
  device0 = jax.local_devices()[0]
  on_tpu = device0.platform == "tpu"
  return jnp.bfloat16 if on_tpu else jnp.float16


dtype_to_names = {
    jnp.bfloat16: ("bf16", "bfloat16"),
    jnp.float16: ("f16", "float16"),
    jnp.float32: ("full", "f32", "float32"),
    jnp.float64: ("f64", "float64"),
}

name_to_dtype = {name: dtype for dtype, names in dtype_to_names.items()  # pylint: disable=g-complex-comprehension
                 for name in names}


def parse_dtype(value: str) -> jnp.dtype:
  """Parses a string representing a dtype into a dtype object."""
  if value == "half":
    return half_dtype()

  try:
    return name_to_dtype[value]
  except KeyError as e:
    raise ValueError(
        f"Unknown dtype '{value}' must be full,half,float16,bfloat16 or a "
        "contraction thereof (e.g. 'f' for 'full', 'bf16' for 'bfloat16')"
    ) from e



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for jmp._src.policy."""

import itertools as it
import unittest

from absl.testing import absltest
from absl.testing import parameterized
import jax
import jax.numpy as jnp
from jmp._src import policy as jmp
import numpy as np

HALF_DTYPES = (np.float16, jnp.float16, jnp.bfloat16)
FULL_DTYPES = (np.float32, jnp.float32)
DTYPES = HALF_DTYPES + FULL_DTYPES
NUMPYS = (np, jnp)


def get_dtype_name(dtype):
  names = {
      np.float16: "float16",
      jnp.bfloat16: "bfloat16",
      np.float32: "float32"
  }
  return names[dtype]


def current_platform():
  return jax.local_devices()[0].platform


def skip_if_unsupported(dtype):
  platform = current_platform()
  if ((platform == "gpu" and dtype == jnp.bfloat16) or
      (platform == "tpu" and dtype in (np.float16, jnp.float16))):
    raise unittest.SkipTest(
        f"{get_dtype_name(dtype)} not supported on {platform}")


class PolicyTest(parameterized.TestCase):

  def assert_dtypes_equal(self, tree_a, tree_b):
    jax.tree_util.tree_map(lambda a, b: self.assertEqual(a.dtype, b.dtype),
                           tree_a, tree_b)

  @parameterized.parameters(*it.product(DTYPES, NUMPYS))
  def test_policy_cast_to_param(self, dtype, np_):
    skip_if_unsupported(dtype)
    policy = jmp.Policy(dtype, dtype, dtype)
    self.assertEqual(policy.param_dtype, dtype)
    tree = {"a": np_.ones([])}
    self.assert_dtypes_equal(policy.cast_to_param(tree),
                             {"a": np_.ones([], dtype)})

  @parameterized.parameters(*it.product(DTYPES, NUMPYS))
  def test_policy_cast_to_compute(self, dtype, np_):
    skip_if_unsupported(dtype)
    policy = jmp.Policy(dtype, dtype, dtype)
    self.assertEqual(policy.compute_dtype, dtype)
    tree = {"a": np_.ones([])}
    self.assert_dtypes_equal(policy.cast_to_compute(tree),
                             {"a": np_.ones([], dtype)})

  @parameterized.parameters(*it.product(DTYPES, NUMPYS))
  def test_policy_cast_to_output(self, dtype, np_):
    skip_if_unsupported(dtype)
    policy = jmp.Policy(dtype, dtype, dtype)
    self.assertEqual(policy.output_dtype, dtype)
    tree = {"a": np_.ones([])}
    self.assert_dtypes_equal(policy.cast_to_output(tree),
                             {"a": np_.ones([], dtype)})

  @parameterized.parameters(*it.product(DTYPES, NUMPYS))
  def test_policy_with_output_dtype(self, dtype, np_):
    policy = jmp.Policy(np_.float32, np_.float32, np_.float32)
    policy = policy.with_output_dtype(dtype)
    self.assertEqual(policy.output_dtype, dtype)

  @parameterized.parameters(("float16", np.float16),
                            ("float32", np.float32),
                            ("bfloat16", jnp.bfloat16))
  def test_get_policy(self, dtype_name, dtype):
    policy = jmp.get_policy(dtype_name)
    self.assertEqual(policy.param_dtype, dtype)
    self.assertEqual(policy.compute_dtype, dtype)
    self.assertEqual(policy.output_dtype, dtype)

  def test_get_policy_almost_dtype(self):
    with self.assertRaisesRegex(ValueError, "Unknown dtype"):
      jmp.get_policy("compute_float16")

  @parameterized.parameters(*it.product(DTYPES, NUMPYS))
  def test_get_policy_mixed(self, dtype, np_):
    full = np_.float32
    policy = jmp.get_policy(f"c={get_dtype_name(dtype)}")
    self.assertEqual(policy.param_dtype, full)
    self.assertEqual(policy.compute_dtype, dtype)
    self.assertEqual(policy.output_dtype, dtype)

  @parameterized.parameters(*it.product(DTYPES, NUMPYS))
  def test_get_policy_compute(self, dtype, np_):
    full = np_.float32
    policy = jmp.get_policy(f"c={get_dtype_name(dtype)},o=full")
    self.assertEqual(policy.param_dtype, full)
    self.assertEqual(policy.compute_dtype, dtype)
    self.assertEqual(policy.output_dtype, full)

  def test_half_dtype(self):
    if current_platform() == "tpu":
      self.assertEqual(jmp.half_dtype(), jnp.bfloat16)
    else:
      self.assertEqual(jmp.half_dtype(), jnp.float16)

  def test_cast_to_full(self):
    half_tree = dict(o=object(),
                     h=jnp.ones([], dtype=jmp.half_dtype()),
                     f=jnp.ones([]),
                     i=jnp.ones([], dtype=jnp.int16))
    full_tree = dict(o=half_tree["o"],
                     h=half_tree["h"].astype(jnp.float32),
                     f=half_tree["f"],
                     i=half_tree["i"])
    self.assertEqual(jmp.cast_to_full(half_tree), full_tree)

  def test_cast_to_half(self):
    dtype = jmp.half_dtype()
    half_tree = dict(o=object(),
                     h=jnp.ones([], dtype=dtype),
                     f=jnp.ones([]),
                     i=jnp.ones([], dtype=jnp.int16))
    full_tree = dict(o=half_tree["o"],
                     h=half_tree["h"],
                     f=half_tree["f"].astype(dtype),
                     i=half_tree["i"])
    self.assertEqual(jmp.cast_to_half(full_tree), half_tree)

  @parameterized.parameters(*it.product(DTYPES))
  def test_str(self, dtype):
    policy = jmp.Policy(dtype, dtype, dtype)
    policy_str = str(policy)
    for str_piece in policy_str.split(","):
      dtype_str = str_piece.split("=")[1]
      self.assertEqual(dtype_str, jmp.dtype_to_names[dtype][0])

if __name__ == "__main__":
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""JMP is a Mixed Precision library for JAX."""

from jmp._src.loss_scale import all_finite
from jmp._src.loss_scale import DynamicLossScale
from jmp._src.loss_scale import LossScale
from jmp._src.loss_scale import NoOpLossScale
from jmp._src.loss_scale import select_tree
from jmp._src.loss_scale import StaticLossScale
from jmp._src.policy import cast_to_full
from jmp._src.policy import cast_to_half
from jmp._src.policy import get_policy
from jmp._src.policy import half_dtype
from jmp._src.policy import Policy

__version__ = "0.0.4"

__all__ = (
    "all_finite",
    "DynamicLossScale",
    "LossScale",
    "NoOpLossScale",
    "select_tree",
    "StaticLossScale",
    "cast_to_full",
    "cast_to_half",
    "get_policy",
    "half_dtype",
    "Policy",
)

#  _________________________________________
# / Please don't use symbols in `_src` they \
# \ are not part of the JMP public API.     /
#  -----------------------------------------
#         \   ^__^
#          \  (oo)\_______
#             (__)\       )\/\
#                 ||----w |
#                 ||     ||
#
try:
  del _src  # pylint: disable=undefined-variable
except NameError:
  pass



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Setup for pip package."""

from setuptools import find_namespace_packages
from setuptools import setup


def _get_version():
  with open('jmp/__init__.py') as fp:
    for line in fp:
      if line.startswith('__version__'):
        g = {}
        exec(line, g)  # pylint: disable=exec-used
        return g['__version__']
    raise ValueError('`__version__` not defined in `jmp/__init__.py`')


def _parse_requirements(requirements_txt_path):
  with open(requirements_txt_path) as fp:
    return fp.read().splitlines()


_VERSION = _get_version()

EXTRA_PACKAGES = {
    'jax': _parse_requirements('requirements-jax.txt'),
}

setup(
    name='jmp',
    version=_VERSION,
    url='https://github.com/deepmind/jmp',
    license='Apache 2.0',
    author='DeepMind',
    description='JMP is a Mixed Precision library for JAX.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author_email='jmp-dev-os@google.com',
    # Contained modules and scripts.
    packages=find_namespace_packages(exclude=['*_test.py']),
    install_requires=_parse_requirements('requirements.txt'),
    extras_require=EXTRA_PACKAGES,
    tests_require=_parse_requirements('requirements-test.txt'),
    requires_python='>=3.8',
    include_package_data=True,
    zip_safe=False,
    # PyPI package information.
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries',
    ],
)
